#ifndef WIN32
#include <sys/time.h>      // for ::gettimeofday()
#endif
#include "mc/manager-mc.h"
#include "mc/session-mc.h"
#include "mc/reply-mc.h"
#include "mc/request-mc.h"
#include "mc/pend-mc.h"
#include "mc/common-mc.h"

static class MCTimeTclClass : public TclClass {
public:
	MCTimeTclClass() : TclClass("MCTime") {}
	TclObject* create(int /*argc*/, const char*const* /*argv*/) {
		return (new MC_Time);
	}

} mctime_class;

int 
MC_Time::command(int argc, const char*const* argv)
{
	Tcl& tcl = Tcl::instance();
	if (!strcmp(argv[1],"get_time")) {
		char wrk[100];
		sprintf(wrk, "%f", gettimeofday());
		tcl.result(wrk);
		return (TCL_OK);
	} 
	return (TclObject::command(argc, argv));
}

double 
MC_Time::gettimeofday(void) const
{
	timeval tv;
	::gettimeofday(&tv, 0);
	return (tv.tv_sec + 1e-6 * tv.tv_usec);
}

static class MCMgrTclClass : public TclClass {
public:
	MCMgrTclClass() : TclClass("MCManager") {}
	
	TclObject* create(int argc, const char*const* /*argv*/) {
		if (argc==4) {
			return (new MC_Manager());
		} else {
			return (TclObject*)NULL;
		}
	}
} mcmgr_class;

/* 
 * Each SRM participant has a MCReceiver for each sender in
 * the group. The job of the MCReceiver is to handle the message
 * when it comes in. In this case, that means caching the document
 * and posting it to the browser.
 */
class MCReceiver : public SRM_PacketHandler {
public:
	MCReceiver(MC_Manager* p, const srm_src& sid) : mcmgr_(p), sid_(sid) {}
	~MCReceiver(void) {}
	virtual void recv(u_char *buf, int len) {
		mcmgr_->recv(buf, len, sid_);
	}
protected:
	MC_Manager* mcmgr_;
	const srm_src& sid_;
};

//virtual
void 
MC_Manager::recv(u_char *pkt, int plen, const srm_src& sid)
{
	dprintf(stderr, "R:");

	sid_ = sid;
	u_int32_t off = ntohl(*(u_int32_t*)&pkt[0]);     // offset into page
	u_int32_t len = ntohl(*(u_int32_t*)&pkt[4]);     // document length
	u_int32_t keylen = ntohs(*(u_int16_t*)&pkt[8]);  // keylength
        u_int32_t hlen = keylen + 10;                    // header length
	dprintf(stderr, "OFF:%d FRAGLEN:%d DOCLEN:%d ", off, plen-hlen, len);

	/*
	 * Not 100% sure what the intention was here so we'll
	 * just comment it out for now since it prevents recv from
	 * being called more than half the time. 
	 */
	// u_int32_t last = hlen + off;
	//if (last > (u_int32_t)plen || last > len) {
	//	/*XXX bump an error counter */
	//	return;
	//}

//#define DROP_MC
#ifdef DROP_MC
	SRM_SourceManager* sm = session_->source_manager();
	SRM_Source* source = sm->consult(sid);
	Tcl& tcl = Tcl::instance();
	// XXX why is source sometimes 0?
	if (source != 0) {
		tcl.evalf("$debugDlg recv %s", source->name());
	}
	dprintf(stderr, "RETVAL: %s\n", tcl.result());
	if (strcmp(tcl.result(), "drop") == 0) {
		dprintf(stderr, "DROP\n");
		return;
	}
#endif // DROP_MC

	char* key = new char[keylen + 1];
	memcpy(key, &pkt[10], keylen);
	key[keylen] = 0;
	dprintf(stderr, "%s\n ", key);
	CacheBuf* cb = lookup_by_key(key);
	if (cb == 0) {
		setup_page(key, len);
		cb = lookup_by_key(key);
	} else if (cb->length() == -1) {
		cb->length(len);
	}
	cb->copyin(off, (char*)&pkt[keylen + 10], plen - hlen);
	Tcl::instance().evalf("%s receive %s %d %d %d",
			      name(), key, off, plen - hlen, len);
	delete[] key;
}

MC_Manager::MC_Manager() :
	SRM_AppMgr(0), session_(0), active_(0), tail_(0)
{
	/* Stores a pointer to a HashEntry object */
	Tcl_InitHashTable(&cache_, TCL_STRING_KEYS);
}

MC_Manager::~MC_Manager()
{
/* XXX need to delete hash table and all its contents */
/* And MCReceivers though we don't have refs to them! */
}

//virtual
SRM_PacketHandler *
MC_Manager::new_source(const srm_src &sid, int /*islocal*/)
{
	return (new MCReceiver(this, sid));
}

/*
 * Fill in the next message ("data unit") to be sent out. 
 * Called by the SRM sending code. 
 */
//virtual
int
MC_Manager::next_ADU(u_char *pkt, int pktlen, srm_src &/*id*/, 
			     int &pkt_type, int & next)
{
	dprintf(stderr,
		"NEXT_ADU: active_:%p next:%p key:%s length:%d offset:%d\n",
		active_, active_->next_, active_->key(), active_->length(), 
		active_->offset());
again:
	int off = active_->offset();
	int len = active_->length();
	int cc = len - off;
	if (cc <= 0) {
		active_ = active_->next_;
		if (active_ != 0)
			dprintf(stderr,
				"AGAIN: active_:%p next:%p key:%s length:%d offset:%d\n",
				active_, active_->next_, active_->key(), active_->length(), 
				active_->offset());

			goto again;
		return (0);
		dprintf(stderr,
			"AGAIN: 0\n");

	}
	int keylen = strlen(active_->key());
	int hlen = keylen + 10;
	if (cc + hlen > pktlen)
		cc = pktlen - hlen;
	*(u_int32_t*)&pkt[0] = htonl(off);
	*(u_int32_t*)&pkt[4] = htonl(len);
	*(u_int16_t*)&pkt[8] = htons(keylen);
	memcpy((char*)&pkt[10], active_->key(), keylen);
	memcpy((char *)&pkt[hlen], active_->data() + off, cc);
	active_->adjust(cc);
	pkt_type = APP_DATA;

//	Tcl& tcl = Tcl::instance();
//	tcl.evalf("mc_trace_d %d %d %s", off, cc, active_->key());	

	int leftover = len - (off + cc);
	if (leftover > 0) {
		leftover += keylen + 10;
		/*XXX*/
		if (leftover > 1000)
			leftover = 1000;
		next = leftover;
	} else if (active_->next_) {
		/*XXX packet size rather than ADU length */
		int len = active_->next_->length();
		if (len > 1000)
			len = 1000;/*XXX*/
		next = len;
	} else {
		active_ = 0;
	}
        return (cc + keylen + 10);
}

void 
MC_Manager::change_key(const char *oldkey, const char *newkey) {
	// XXX use a delete method
	Tcl_HashEntry* he = Tcl_FindHashEntry(&cache_, oldkey);
	CacheBuf* cb = (CacheBuf*)Tcl_GetHashValue(he);	
	Tcl_DeleteHashEntry(he); 
	cb->change_key(newkey);
	int newEnt;
	he = Tcl_CreateHashEntry(&cache_, newkey, &newEnt);
	Tcl_SetHashValue(he, (char*)cb);
}

void 
MC_Manager::enter(const char* key, const char* buffer, int len)
{
	CacheBuf* cb = create(key);
	cb->set(buffer, len);
}

CacheBuf* 
MC_Manager::create(const char* key)
{
	int newEnt;
	Tcl_HashEntry* he = Tcl_CreateHashEntry(&cache_, (char*)key,
						(int*)&newEnt);
	if (!newEnt) {
		/* XXX - shouldn't happen.  delete the old entry */
		CacheBuf* cb = (CacheBuf*)Tcl_GetHashValue(he);
		delete cb;
	}
	CacheBuf *cb = new CacheBuf(key);
	Tcl_SetHashValue(he, (char*)cb);
	return cb;
}

CacheBuf* 
MC_Manager::lookup_by_key(const char* key)
{
	Tcl_HashEntry* he = Tcl_FindHashEntry(&cache_, (char*)key);
	//dprintf(stderr, "lookup_by_key: HE:%d KEY:%s ", (int)he, key);
	if (he != 0) {
		CacheBuf *cb =(CacheBuf*)Tcl_GetHashValue(he); 
		//dprintf(stderr, "OLD:%p \n", cb);		
		return cb;
	}
	//dprintf(stderr, "NO ENTRY\n");		
	return (0);
}

CacheBuf *
MC_Manager::setup_page(const char *key, int len) 
{
	CacheBuf *cb = create(key);

//	Tcl& tcl = Tcl::instance();
//	tcl.evalf("mc_trace_ps %d %s", len, key);	

	cb->length(len);
	return cb;
}

void 
MC_Manager::start(const char* key)
{
	
	Tcl_HashEntry* he = Tcl_FindHashEntry(&cache_, (char*)key);
	if (he == 0)
		/*XXX shouldn't happen */
		return;


	CacheBuf* cb = (CacheBuf*)Tcl_GetHashValue(he);

	/*
	 * If we're already streaming packets, add this buffer
	 * to the tail of the queue.  Otherwise, activate
	 * the streaming process.
	 */
	cb->next_ = 0;
	if (active_) {
		tail_->next_ = cb;
		tail_ = cb;
	} else {
		active_ = cb;
		tail_ = cb;
		active_->start();
		/*
		 * XXX SRM question - why do we have to give
		 * the byte count?  XXX also, this is wrong
		 * since it wants the next packet size
		 * not the next chunk size...
		 */
		int len = active_->length();
		if (len  > 1000)
			len = 1000;/*XXX*/
		session_->begin_xmit(len);
	}
}

int 
MC_Manager::command(int argc, const char*const* argv)
{
	Tcl& tcl = Tcl::instance();
	if (!strcmp(argv[1],"attach_session")) {
		MC_Session* p = (MC_Session*)TclObject::lookup(argv[2]);
		if (p == 0) {
			tcl.result("Couldn't find session");
			return TCL_ERROR;
		}
		session_ = p;
		return (TCL_OK);
	}
	if (!strcmp(argv[1],"change_key")) {
		const char *oldkey = argv[2];
		const char *newkey = argv[3];
		change_key(oldkey, newkey);
		return (TCL_OK);
	}
	if (!strcmp(argv[1], "length")) {
		CacheBuf* cb = lookup_by_key(argv[2]);
		if (cb == 0) {
			tcl.result("no key");
		} else {
			char wrk[20];
			sprintf(wrk, "%d", cb->length());
			tcl.result(wrk);
		}
		return (TCL_OK);
	}
	if (!strcmp(argv[1], "lookup")) {
		CacheBuf* cb = lookup_by_key(argv[2]);
		if (cb == 0 || cb->data() == 0) {
			return (TCL_OK);
		}

		char* prefix = cb->data();
		int len = cb->length();
		int plen = strlen(prefix) + 1;
		char* data = prefix + plen;
		len -= plen;

		/*
		 * XXX hack - the data is passed in a buffer of binary
		 * data stored in the global variable buffer$self.
		 */
		char wrk[20];
		sprintf(wrk, "buffer%s", name());
		Tcl_Obj* id = Tcl_NewStringObj(wrk, strlen(wrk));
		/*XXX maybe we should save the Tcl_Obj in CacheBuf */
		Tcl_Obj* p = Tcl_NewStringObj(data, len);

		(void)Tcl_ObjSetVar2(tcl.interp(), id, (Tcl_Obj *)0, p,
				     TCL_GLOBAL_ONLY);
		tcl.result(prefix);
		return (TCL_OK);
	}
			
	/*
	 * <otcl> MultiCache repair_hole { key }
	 * XXX comment
	 */
	if (!strcmp(argv[1], "send_request")) {
		char *key = (char *)argv[2];
		MC_Request *request = new MC_Request(this, key);
		// find the request and set it
		CacheBuf* cb = lookup_by_key(key);		
		if (cb == 0) {
			/* we've never seen this doc before; send
			* -1 as its length because we don't know it now */
			cb = setup_page(key, -1);
		}
		cb->request(request);
		session_->sched_request(request, sid_);
		dprintf(stderr, "REQUEST SCHEDULED FOR: %s\n", key);
		return (TCL_OK);
	}

	/*
	 * <otcl> MultiCache send_pending { key }
	 * XXX comment
	 */
	if (!strcmp(argv[1], "send_pending")) {
		char *key = (char *)argv[2];
		char *http_server = (char*)argv[3];
		MC_Pending *pending = new MC_Pending(this, key, http_server);
		session_->sched_pending(pending, sid_);
		dprintf(stderr, "PENDING SCHEDULED FOR: %s\n", key);
		return (TCL_OK);
	}

	/*
	 * <otcl> MultiCache send_reply { key }
	 * XXX comment
	 */
	if (!strcmp(argv[1], "send_reply")) {
		char *key = (char *)argv[2];		
		MC_Reply *reply = new MC_Reply(this, key);
		session_->sched_reply(reply, sid_);
		dprintf(stderr, "REPLY SCHEDULED FOR: %s\n", key);
		return (TCL_OK);
	}

	/*
	 * <otcl> MultiCache insert { key data len }
	 * Insert the data with the corresponding key
	 * into the underlying reliable multicast session.
	 * Causes the data to be sent over the network.
	 * XXX data/len should perhaps be replaced with a bufferID.
	 * XXX newer hack - buffer pass in global var
	 */
	if (!strcmp(argv[1], "insert")) {
		if (argc != 5) {
			tcl.result("wrong number of args to send");
			return (TCL_ERROR);
		}
		const char* key = argv[2];
		int len = atoi(argv[3]);

		/*XXX hack - a hook to put some prefixed text */
		const char* prefix = argv[4];
		/*
		 * XXX hack - the data is passed in a buffer of binary
		 * data stored in the global variable buffer$self.
		 */
		char wrk[20];
		sprintf(wrk, "buffer%s", name() );
		Tcl_Obj* id = Tcl_NewStringObj(wrk, strlen(wrk));
		Tcl_Obj* p = Tcl_ObjGetVar2(tcl.interp(), id, (Tcl_Obj *)0,
					    TCL_LEAVE_ERR_MSG|TCL_GLOBAL_ONLY);
		if (p == 0)
			/*XXX tcl should set up this var */
			abort();

		int slen;
		char* data = Tcl_GetStringFromObj(p, &slen);
		if (len != slen)
			abort();
		/*
		 * Enter the page into the cache, then
		 * kick it out to the session. The data we enter includes
		 * the prefix from argv[4] and the data from the global
		 * variable buffer$self.
		 */

		CacheBuf* cb = create(key);
		int plen = strlen(prefix);
		cb->length(len + plen + 1);
		cb->copyin(0, prefix, plen + 1);
		cb->copyin(plen + 1, data, len);
		start(key);

		/*
		 * Free objects.
		 */
		Tcl_DecrRefCount(id);
		Tcl_DecrRefCount(p);

		/* 
		 * XXX not sure this is really necessary. All we're trying to 
		 * do is make sure that the tcl side of the cached data struc-
		 * ture is set up. However, the Holes structure should always
		 * be empty since we always have all of the data here. Is
		 * it necessary to set up the empty data structure? Yes,
		 * because we check the holes_ database on the tcl side
		 * in the tcl handle_request method to see if we have
		 * *all* of the data. An alternative would be to delete the
		 * holes_ structure for any page that has been fully received.
		 * 
		 * receive { url off datalen pagelen }
		 */
		Tcl::instance().evalf("%s receive %s 0 %d %d",
				      name(), key, cb->length(), cb->length());

		return (TCL_OK);
	}
	return (TclObject::command(argc, argv));
}

//virtual 
void 
MC_Manager::handle_request(const srm_src &sid, u_char *pb, int /*len*/) {

	/* We end up using this before the current event is over. */
	sid_ = sid;

	u_int16_t keylen = ntohs(*(u_int16_t*)&pb[0]);  
	u_int16_t numrps = ntohs(*(u_int16_t*)&pb[2]);  
        RequestPair* rp = (RequestPair*)&pb[4];

	char *key = new char[keylen + 1];	    
	memcpy(key, (char*)(rp + numrps), keylen);
	key[keylen] = 0;
	dprintf(stderr, "HANDLE_REQUEST %s", key);

	/* XXX make this more conservative */
	CacheBuf *cb = lookup_by_key(key);
	if (cb != 0) {
		MC_Request *request = cb->request();
		if (request != 0) {
			request->backoff();
		}	
	}

	/* backoff the requests & schedule repairs */
	unsigned int i;
	for (i = 0; i < numrps; i++) {
		/* might use Tcl_Merge and just call evalf() once */
		u_int32_t offset = ntohl(rp[i].offset());
		u_int32_t len = ntohl(rp[i].length());
		dprintf(stderr, "\tRP:%d %d\n", offset, len);
		Tcl::instance().evalf("%s check_reply %s %d %d",
				      name(), key, offset, len);
	}	
}

void 
MC_Manager::handle_reply(const srm_src &sid, u_char *pkt, int plen) 
{
	u_int32_t len = ntohl(*(u_int32_t*)&pkt[4]);     // document length
	if (len == 0) {
		handle_pending(sid, pkt, plen);
	} else {
		recv(pkt, plen, sid);
	}
}

void 
MC_Manager::handle_pending(const srm_src &/*sid*/, u_char *pkt, int /*plen*/) 
{
	u_int32_t keylen = ntohs(*(u_int16_t*)&pkt[8]);  // keylength

	char* key = new char[keylen + 1];
	memcpy(key, &pkt[10], keylen);
	key[keylen] = 0;
	dprintf(stderr, "HANDLE_PENDING:%s\n ", key);

	Tcl& tcl = Tcl::instance();
	tcl.evalf("%s cancel_pending %s", name(), key);

	delete[] key;
}






