import Application StockcastUI WebPoller/CheckFree \
	AnnounceListenManager/AS/Client/CASA

#
# StockCaster is a stock quotes dissemination application using SRMv2
# Author: Tina Wong
# Creation Date: 4/20/99
# Status: pre-alpha
#
Class StockcastApp -superclass Application

#
# Constructor
#
StockcastApp public init { argv } {
	$self next stockcast

	# parse arguments
	set o [$self options]
	$o register_boolean_option -recv recv
	$o register_boolean_option -src src
	set netspec [$o parse_args $argv]
	global ssd
	if { [$self get_option recv] != "" } {	
		set ssd(ctype) "recv"
	} elseif { [$self get_option src] != "" } {
		set ssd(ctype) "src"		
	} else {
		puts stderr "stockcast: must specify as recv or src"
		$self fatal
	}

	if { $netspec == "" } {
		puts stderr "stockcast: [-recv|-src] address"
		$self fatal "unknown/invalid address"
	}

	# srmv2 related variables
	# session_ and source_ are arrays
	$self instvar session_ source_ port_ ttl_
	set ab [new AddressBlock $netspec]
	set addr [$ab addr]
	set port_ [$ab sport]
	set ttl_ [$ab ttl]
	delete $ab
	
	set session_($addr) [srm_create_session $addr $port_ $port_ $ttl_]
	set source_($addr) [srm_create_source $session_($addr)]
	srm_callbacks $session_($addr) srm_recv srm_should_recover srm_read_adu \
		srm_source_update srm_recv_cid

	# init mappings
	$self instvar addr_names_
	foreach sym $ssd(pref) {
		set addr_names_($sym) $addr
	}

	# as1 client
	$self instvar as_
	global params ssd
	set as_ [new AnnounceListenManager/AS/Client/CASA \
		$self [array get params] [array get ssd]]

	# build user interface
	if { [$self get_option recv] != "" } {
		$self instvar ui_
		set w .top$self
		frame $w
		set ui_ [new StockcastUI $w $self $as_]
		pack $w
		
	# configure webpoller
	} else {
		wm withdraw .
		$self instvar wp_
		set wp_ [new WebPoller/CheckFree $self]
	}

	# start announcing data to servent
	$as_ start 
}

StockcastApp public recv_mappings { orphans mappings } {
	$self instvar addr_names_ 

	set addr_names_(orphans) $orphans
	if { $mappings != "" } {
		array set addr_names_ $mappings
		$self regroup
	}
}

StockcastApp public recv_bwalloc { bwalloc } {
	$self instvar timer_names_ clearbit_names_ wp_

	set data_size [$wp_ get_data_size]
	array set bw_names $bwalloc
	foreach sym [array names bw_names] {
		set period [expr double($data_size) / double($bw_names($sym))]		
		if ![info exists timer_names_($sym)] {
			set timer_names_($sym) [new StockcastSendTimer $self $period $sym]
			set clearbit_names_($sym) 1
		} else {
			set timer_names_($sym) set period_ $period
		}
	}
}

StockcastApp public clear { sym } {
	$self instvar clearbit_names_
	set clearbit_names_($sym) 1
}

#
# Join and leave multicast address depending on user interest
# and mapping information.
#
StockcastApp public regroup { } {
	$self instvar session_ source_ ui_ addr_names_

	set subscribed [array names session_]
	
	# receiver only interested in those symbols specified in ui
	if { [$self get_option recv] != "" } {		
		set interested [$ui_ get_interest]

	# source send all symbols it is responsible for
	} elseif { [$self get_option src] != "" } {
		set interested [array names addr_names_]
		
		# notify webpoller to get new symbols
		$self instvar wp_
		set new [ldiff $interested [$wp_ set symbols_]]

		# XXX need to check whether symbols is in subset responsible,
		# if doing parallel webpollling (e.g. alphabetical)		
		if { $new != "" } {
			$wp_ add_symbols $new
		}
	}
	
	# join mcast groups if not already in	
	foreach sym $interested {
		# if no one show interest to sym before, need to wait til reflected in mapping
		if [info exists addr_names_($sym)] {
			set addr $addr_names_($sym)
			if { [lsearch $subscribed $addr] == -1 } {
				$self instvar port_ ttl_

				set session_($addr) [srm_create_session $addr $port_ $port_ $ttl_]
				set source_($addr) [srm_create_source $session_($addr)]
				srm_callbacks $session_($addr) srm_recv srm_should_recover \
					 srm_read_adu srm_source_update srm_recv_cid

				puts stderr "joining: $addr $session_($addr) $source_($addr)"
			}
		}
	}

	# leave mcast groups not interested anymore
	foreach addr [array names session_] {
		set syms [$self addr_2_names $addr]
		if { [lintersect $syms $interested] == "" } {
			#srm_destroy_session $session_($addr)
			unset session_($addr)
			unset source_($addr)
			
			puts stderr "leaving: $addr"			
		}
	}
}

#
# Given an address, find the current set of symbols associated with it
#
StockcastApp private addr_2_names { addr } {
	$self instvar addr_names_

	foreach sym [array names addr_names_] {
		if { "$addr_names_($sym)" == "$addr" } {
			lappend all $sym
		}
	}
	if { ![info exists all] } {
		puts stderr "stockcast: addr $addr not mapped"
		return ""
	} else {
		return $all
	}
}

#
# SRM related methods 
#

#
# Called by srm if someone needs to recover the specified data.
#
StockcastApp public read_data { source cid seqno } {
	$self instvar index_

	set sym [$self cid_2_name $source $cid]
	
	if { $sym == "" || ![info exists index_($sym)] } {
		return ""
	} else {
		return $index_($sym)
	}
}


#
# Called by webpoller when there is data to send.
#
StockcastApp public send_data { sym data } {
	$self instvar source_ index_ cid_names_ addr_names_ clearbit_names_

	# orphan symbol
	if [info exists clearbit_names_($sym)] {
		
	# need to wait
	} elseif { $clearbit_names_($sym) == 0 } {
		return
	}
	set clearbit_names_($sym) 0

	# no interest for this sym, use orphans addr
	if { ![info exist addr_names_($sym)] } {
		set addr $addr_names_(orphans)
	} else {
		set addr $addr_names_($sym)
	}

	# allocate a cid for this sym if one does not exist
	set p [$self name_2_cid $sym]
	set source [lindex $p 0]
	set cid [lindex $p 1]
	if { $cid == "" || $source != $source_($addr)} {
		set cid [srm_calloc $source_($addr) 0 $sym]
		set cid_names_($source_($addr),$cid) $sym
	}
	set index_($sym) $data

	# send the data to the session associated with the sym
	set seqno [srm_send $source_($addr) $cid $data]
	puts stderr "$addr $cid $seqno"
}


#
# Called when received data from the session.
#
StockcastApp public recv_data { source cid seqno data } {
	$self instvar index_ ui_

	# do a reverse mapping lookup to get sym
	set sym [$self cid_2_name $source $cid]

	# store the data onto local memory
	set index_($sym) $data

	# display the data is there is interest in sym
	if {[$self get_option recv] != "" && [$ui_ is_interest $sym] == 1} {
		$ui_ fillrow_symbol $sym $data
	}
}

#
# Default behavior is to recover only when interset in symbol.
# Subclasses can change this reliability strategy.
#
StockcastApp public should_recover { source cid sseq eseq } {
	$self instvar ui_
	
	# do a reverse mapping lookup to get sym
	set sym [$self cid_2_name $source $cid]
	if {[$ui_ is_interest $sym] == 1} {	
		return 1
	} else {
		return 0
	}
}


#
# Receive an sym to SRM mapping information.
#
StockcastApp public recv_cid { source cid parent_cid name } {
	$self instvar cid_names_
	set cid_names_($source,$cid) $name
}


#
# Given a cid, this method returns the application level
# name (sym) associated with it.
#
StockcastApp public cid_2_name { source cid } {
	$self instvar cid_names_
	
	if [info exists cid_names_($source,$cid)] {
		set sym $cid_names_($source,$cid)
	} else {
		set sym [srm_get_container_name $source $cid]
	}
	return $sym
}

#
# Give a name (sym), this method returns the (source,cid) pair
# associated with it or an empty if it does not exist.
#
StockcastApp public name_2_cid { sym } {
	$self instvar cid_names_

	foreach p [array names cid_names_] {
		if { "$cid_names_($p)" == "$sym" } {
			return [split $p ,]
		}
	}
	return ""
}

#
# SRMv2 tcl callbacks
#

proc srm_recv { src cid seqno data } {
	global stockcast
	
	$stockcast recv_data $src $cid $seqno $data
}

proc srm_read_adu { src cid seqno } {
	global stockcast
	
	return [$stockcast read_data $src $cid $seqno]
}

proc srm_should_recover { src cid sseq eseq } {
	global stockcast
	$stockcast should_recover $src $cid $sseq $eseq
}


proc srm_recv_cid { src cid parent_cid name } {
	global stockcast
	$stockcast recv_cid $src $cid $parent_cid $name
}

proc srm_source_update { src info } {
	# do nothing now
}

