#
# Copyright (c) 1998 Regents of the University of California.
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. All advertising materials mentioning features or use of this software
#    must display the following acknowledgement:
# 	This product includes software developed by the MASH Research
# 	Group at the University of California Berkeley.
# 4. Neither the name of the University nor of the Research Group may be
#    used to endorse or promote products derived from this software without
#    specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#
# $Header: /usr/src/mash/repository/mash/mash-1/tcl/mc/holes.tcl,v 1.9 1998/06/28 17:30:12 kwright Exp $

Class Holes 

Holes instproc init {length} {
	$self add 0 $length
}

Holes instproc destroy args {
	$self next
}
# If we have an unknown length (only
# one hole and it's offset is 0 and it's 
# length is -1), update
# it. 
Holes instproc update_length {doclen} {
	$self instvar list_
	set offsets [array names list_]
	if {[llength $offsets] == 1} {
		if {[lsearch $offsets 0] != -1} {
			if {$list_(0) == -1} {
				unset list_(0)
				$self add 0 $doclen
			}
		}
	}
}

# Prints the [offset,length] pairs
Holes instproc oldPrint {} {
	$self instvar list_
	set keys [array names list_]
	foreach key $keys {
		puts "$key $list_($key)"
	}
}

# Return the array as a list:
# <offset><len><offset><len>...
Holes instproc listize {} {
	$self instvar list_
	set offsets [array names list_]
	set len [llength $offsets]
	set offsets [lsort -integer -increasing $offsets]
	if {[llength $offsets] == 0} {
		return ""
	} 
	foreach offset $offsets {
		lappend answer $offset $list_($offset)
	}
	return $answer
}

# Copy the Holes object in the argument to a new object
Holes instproc copy {} {
	$self instvar list_
	# length doesn't matter because we fill it right away
	set newh [new Holes 4]
	$newh fill 0 4
	set offsets [array names list_]
	if {[llength $offsets] == 0} {
		return $newh
	} 
	foreach offset $offsets {
		$newh add $offset $list_($offset)
	}
	return $newh
}

# Prints the [offset, endOffset] pairs
Holes instproc print {} {
	$self instvar list_
	set keys [array names list_]
	set keys [lsort -integer -increasing $keys]
	foreach key $keys {
		if {$list_($key) == -1} {
			puts "      $key $list_($key)"
			continue
		}
		set endOffset [expr $key+$list_($key) - 1]
		puts stderr "      $key $endOffset"
	}
}

Holes instproc numHoles {} {
	$self instvar list_
	array size list_
}

# Because it's an error to request
# an element that doesn't exist
Holes instproc find key {
	$self instvar list_
	set keys [array names list_]
	set answer [lsearch $keys $key]
	if {$answer == -1} {
		return ""
	} 
	return $list_($key)
}

Holes instproc remove key {
	$self instvar list_
	if {[$self find $key] != ""} {
		unset list_($key)
	}
}

# Don't allow a 0 length 
Holes instproc add {key data} {
	$self instvar list_
	if {$data == 0} {
		return ""
	}
	set list_($key) $data
}

# Don't allow a 0 length 
Holes instproc setLength {offset length} {
	$self instvar list_
	if {[$self find $offset] == ""} {
		return ""
	}
	if {$length == 0} {
		$self remove $offset
	} else {
		set list_($offset) $length
	}
}

Holes instproc min {x y} {if {$x < $y} {return $x} else {return $y} }
Holes instproc max {x y} {if {$x > $y} {return $x} else {return $y} }

Holes instproc dig {offset length} {
	$self instvar list_
	set offsets [array names list_]
	
	# cover the case of no holes 
	if {[$self numHoles] == 0} {
		$self add $offset $length
		return
	}

	# cover whole-doc requests	
	if {$length == -1} {
		foreach holeLeft $offsets {
			$self remove $holeLeft
		}
		$self add 0 -1
		return
	}

	set digLeft $offset
	set digRight [expr $offset + $length - 1]
	set offsets [array names list_]

	set overlap 0
	foreach holeLeft $offsets {
		set holeRight [expr $holeLeft + $list_($holeLeft) -1]
#		puts "holeLeft:$holeLeft holeRight:$holeRight digLeft:$digLeft digRight:$digRight"
		if { $holeRight+1 < $digLeft || $holeLeft-1 > $digRight } {
			# no overlap
			continue
		} else {
			# overlap
			set newLeft [$self min $holeLeft $digLeft]
			set newRight [$self max $holeRight $digRight]
			$self remove $holeLeft
			$self remove $digLeft
			$self add $newLeft [expr $newRight - $newLeft + 1]
			# we merged so reset $digLeft and $digRight. this
			# allows us to merge correctly with other blocks
			set digLeft $newLeft
			set digRight $newRight
			set overlap 1
		}
	}
	if {$overlap == 0} {
		# haven't added hole yet because it didn't overlap
		$self add $offset $length
	}
}

Holes instproc fill {offset length} {
	$self instvar list_
	set offsets [array names list_]

	# cover the "whole-doc" case: where we fill in 
	# the whole document by specifying <0 -1>
	if {$length == -1} {
		foreach holeLeft $offsets {
			$self remove $holeLeft
		}
		return 0
	}

	set recvLeft $offset
	set recvRight [expr $offset + $length - 1]

	foreach holeLeft $offsets {
		set holeRight [expr $holeLeft + $list_($holeLeft) - 1]
		if {$holeRight < $recvLeft || $recvRight < $holeLeft } {
			# no overlap
			# hole: |---|        or        |---|
			# recv:       |---|       |---|
			continue

		} elseif {$holeRight == $recvRight && $holeLeft == $recvLeft} {
			# hole: |-------|
			# recv: |-------|
			$self remove $holeLeft

		} elseif {$holeLeft < $recvLeft && $holeRight < $recvRight} {
			# no-overlap covered above so we can assume they overlp
			# hole: |-------|          
			# recv:      |-------|
			set list_($holeLeft) [expr $recvLeft - $holeLeft]

		} elseif {$recvLeft < $holeLeft && $recvRight < $holeRight} {
			# hole:      |-------|
			# recv: |-------|
			set newOffset [expr $recvRight + 1]
			set newLength [expr $holeRight - $recvRight]
			$self remove $holeLeft
			$self add $newOffset $newLength

		} elseif {$holeLeft <= $recvLeft && $recvRight <= $holeRight} {
			# hole: |-------| or |-------| or |-------|
			# recv:   |--|       |--|              |--|
#puts "holeLeft:$holeLeft holeRight:$holeRight recvLeft:$recvLeft recvRight:$recvRight"
			$self setLength $holeLeft [expr $recvLeft - $holeLeft]
			set newOffset [expr $recvRight + 1]
			set newLength [expr $holeRight - $recvRight]
			$self add $newOffset $newLength

		} elseif {$recvLeft <= $holeLeft && $holeRight <= $recvRight} {
			# hole:   |--|    or |--|      or      |--|
			# recv: |-------|    |-------|    |-------|
			$self remove $holeLeft

		} else {
#			puts "holes.tcl error"
		}
	}
	return [$self numHoles]
}

# return 1 if we have this data in its entirety
# return 0 otherwise
# XXX bug if we ask for offset past length of data 
Holes instproc have_data {left len} {
	$self instvar list_

	# cover the case of a whole-doc request;
	# invariant: if a Holes structure even exists,
	#  we don't have all the data. 
	if { $len == -1 } {
		return 0
	}

	set right [expr $left + $len - 1]
	set offsets [array names list_]
	foreach holeLeft $offsets {
		set holeRight [expr $holeLeft+$list_($holeLeft) - 1]
		if { $holeRight < $left || $holeLeft > $right } {
			# no overlap
			continue
		} else {
			return 0
		}
	}
	return 1
}

# Delete & return the first hole
Holes instproc remove_first {} {
	$self instvar list_

	# cover the case of no holes
	if {[$self numHoles] == 0} {
#puts "REMOVE_FIRST: 0 length Hole"
		return -1
	}
	set offsets [array names list_]
	set offsets [lsort -integer -increasing $offsets]	
	lappend answer [lindex $offsets 0]
	lappend answer $list_($answer)
	return $answer
}


proc testListize {} {
	Holes h 100
	h add 0 10
	h add 20 30
	h add 40 50
	set mylist [h listize]
	puts "e 0 10 20 30 40 50"
	puts $mylist
}

proc testRemoveFirst {} {
	Holes h 20
	h fill 0 20
	set answer [h remove_first]
	puts "e -1"
	puts $answer
	h dig 2 5
	h dig 8 7
	puts "e 2 6, 8 14" 
	h print
	set answer [h remove_first]
	puts "e 2 5"
	puts $answer
}

proc testHaveData {} {
	Holes h 20
	puts "e 0"
	set answer [h have_data 3 10]
	puts $answer
	puts "e 0"	
	set answer [h have_data 1 1]
	puts $answer
	Holes h 30
	h fill 0 6
	h fill 15 6
	puts "e 6 14, 21 29"
	h print
	set answer [h have_data 1 3]
	puts "e 1"
	puts $answer
	set answer [h have_data 4 2]
	puts "e 6 14, 21 29"
	h print
	puts "e 1"
	puts $answer
	set answer [h have_data 4 3]
	puts "e 0"
	puts $answer
	set answer [h have_data 6 2]
	puts "e 0"
	puts $answer
	set answer [h have_data 13 2]
	puts "e 0"
	puts $answer
	set answer [h have_data 13 3]
	puts "e 0"
	puts $answer
	set answer [h have_data 15 6]
	puts "e 1"
	puts $answer
	set answer [h have_data 18 3]
	puts "e 1"
	puts $answer
	set answer [h have_data 21 9]
	puts "e 0"
	puts $answer
}	

proc testFill {} {
	Holes h 20
	puts "e 0 20"
	h oldPrint
	# no overlap
	# hole: |---|        
	# recv:       |---|  
	h fill 50 10
	puts "e 0 20"
	h oldPrint
	# hole: |-------|
	# recv: |-------|
	h fill 0 20
	puts "e nothing"
	h oldPrint
	# hole:       |---|
	# recv: |---|
	h add 20 10
	h fill 0 20
	puts "e 20 10"
	h oldPrint
	# hole: |-------|
	# recv:      |-------|
	h fill 20 10
	h add 0 20
	h fill 15 6
	puts "e 0 15"
	h oldPrint
	# hole:      |-------|
	# recv: |-------|
	h fill 0 15
	h add 15 6
	h fill 11 7
	puts "e 18 3"
	h oldPrint
	# hole: |-------|
	# recv:   |--|
	h fill 18 3
	h add 0 20
	h fill 11 4
	puts "e 0 11 & 15 5"
	h oldPrint
	# hole:   |--|
	# recv: |-------|
	h fill 0 11
	h fill 15 5
	h add 2 5
	h fill 0 11
	puts "e nothing"
	h oldPrint
	# hole: |----|
	# recv:      |----|
	h add 14 4
	h fill 17 3
	puts "e 14 3"
	h oldPrint
	# hole:      |----|
	# recv: |----|
	h fill 14 3
	h add 21 7
	h fill 20 2
	puts "e 22 6"
	h oldPrint
	# hole: |-------|
	# recv: |--|
	h fill 22 6
	h add 12 6
	h fill 12 2
	puts "e 14 4"
	h oldPrint
	# hole: |-------|
	# recv:      |--|
	h fill 15 3
	puts "e 14 1"
	h oldPrint
	# hole: |--|   
	# recv: |-------|
	h fill 14 1
	h add 7 9
	h fill 7 11
	puts "e nothing"
	# hole:      |--|   
	# recv: |-------|
	h add 7 9
	h fill 2 14
	puts "e nothing"
	puts DONE
}

# Note: I think the expected values are off by one (too long)
proc shortTest {} {
	Holes h 2532
	puts "e 0 2532"
	h print
	h fill 0 971
	puts "e 971 2532"
	h print
	h fill 971 971
	puts "e 1942 2532"
	h print
	h fill 1942 590
	puts "e nothing"
	h print
	puts DONE
}

proc testDig {} {
	Holes h 40
	puts "e 0 39"
	h print
	h fill 0 40
	puts "e nothing"
	h print
	# hole:   
	# dig:  |--|   
	h dig 0 2
	puts "e 0 1"
	h print
	# hole: |--|   
	# dig:  |-------|
	h dig 0 40
	puts "e 0 39"
	h print
	h fill 0 10
	puts "e 10 39"
	h print
	h fill 20 10
	puts "e 10 19,30 39"
	h print
	# hole:      |----|
	# dig:  |----|
	h dig 8 2
	puts "e 8 19, 30 39"
	h print
	# hole: |----|
	# dig:       |----|
	h dig 20 5
	puts "e 8 24, 30 39"
	h print
	# no overlap
	# hole:          |---|        
	# dig:    |---|  
	h dig 3 4
	puts "e 3 6, 8 24, 30 39"
	h print
	# hole: |-------|
	# recv: |-------|
	h fill 8 20
	puts "e 3 6, 30 39"
	h print
	# no overlap
	# hole: |---|        
	# dig:         |---|  
	h dig 8 20
	puts "e 3 6, 8 27, 30 39"
	h print
	# hole:      |-------|
	# dig: |-------|
	h fill 8 20
	h dig 15 17
	puts "e 3 6, 15 39"
	h print
	# hole: |-------|
	# dig:       |-------|
	h dig 4 5
	puts "e 3 8, 15 39"
	h print
	# hole: |-------|
	# dig:     |--|
	h dig 30 5
	puts "e 3 8, 15 39"
	h print
	# hole:   |--|
	# dig:  |-------|
	h dig 2 8
	puts "e 2 9, 15 39"
	h print
	# hole:      |----|
	# dig:  |----|
	h dig 11 6
	puts "e 2 9, 11 39"
	h print
	# hole:      |----|
	# dig:  |----|
	h fill 11 20
	puts "e 2 9, 31 39"
	h print
	h dig 15 17
	puts "e 2 9, 15 39"
	h print
	# hole: |-------|
	# dig:  |--|
	h dig 15 4
	puts "e 2 9, 15 39"
	h print
	# hole: |-------|
	# dig:       |--|
	h fill 36 4
	puts "e 2 9, 15 35"
	h print
	h dig 30 6
	puts "e 2 9, 15 35"	
	h print
	# hole:      |--|   
	# dig:  |-------|
	h dig 10 6
	puts "e 2 35"
	h print
	puts DONE
}

proc testCopy {} {
	Holes h 40
	h fill 0 100
	set newh [h copy]
	puts "e nothing"
	$newh print
	h add 1 5
	h add 10 20
	h add 300 300
	set newh [h copy]
	puts "e 1 5, 10 29, 300 599"
	$newh print
}

proc testUpdateLength {} {
	Holes h -1
	puts "e 0 -1" 
	h print
	h update_length 30
	puts "e 0 29" 
	h print
	delete h
	Holes h 30
	puts "e 0 29"
	h print
	h update_length 30
	puts "e 0 29"
	h print
	delete h
	Holes h 30
	puts "e 0 29"
	h print
	h fill 0 10
	puts "e 10 29"
	h print
	h update_length 30
	puts "e 10 29"
	h print
}


