#
# Copyright (c) 1997 Regents of the University of California.
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. All advertising materials mentioning features or use of this software
#    must display the following acknowledgement:
# 	This product includes software developed by the MASH Research
# 	Group at the University of California Berkeley.
# 4. Neither the name of the University nor of the Research Group may be
#    used to endorse or promote products derived from this software without
#    specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#
# @(#) $Header: /usr/src/mash/repository/mash/mash-1/tcl/mc/server.tcl,v 1.27 1998/06/28 17:30:14 kwright Exp $
#
# 
# server.tcl - Create a TCPServer for HTTP protocol
#

import TCP Observable

# A class that acts as a middleman for web requests
# and manipulates the local browser with "netscape -remote"...
# this needs to be decoupled.
Class WebProxy -superclass { TCP/Server Observable }

#XXX
source ../tcl8.0/library/http2.0/http.tcl
# to autoload the ::http namespace
::http::formatQuery sdsds

Class ProxyConnection -superclass TCP

WebProxy instproc create_channel chan {
	return [new ProxyConnection [$self set webcache_] $self $chan]
}

ProxyConnection public init { wc observer chan } {
	$self instvar webcache_ observer_  launcher_ 
	set webcache_ $wc
	set observer_ $observer
	set launcher_ [new Launcher $wc]
	$self open $chan
}

WebProxy instproc init { port webcache } {
	$self next
	$self instvar webcache_ 
	set webcache_ $webcache
	$self open $port 
}

ProxyConnection instproc recv s {

	set words [split $s]
	if { [lindex $words 0] == "GET" } {
		#XXX check version
		set url [string trim [lindex $words 1]]
		mc_trace_ps -1 $url

		$self instvar webcache_ data_ trace_ observer_
		#
		# If a trace object is attached, then dispatch
		# the url to it... (should this merely be an observer?)
		$observer_ notify_observers trace_url $url

		# Are we using the cache? XXX I just added this to time fetches
		# when not using the cache. 
		global mc_nocache
		if {$mc_nocache == "yes"} {
			$self direct_launch $url
			return
		}

		# First see if it's in the srm cache
		set prefix [$webcache_ lookup $url]
		if { [$webcache_ have_all_data $url] == "yes" } {
			#
			# We hit.  XXX lookup_url returns the
			# the prefix we installed (i.e., the content
			# type) and puts the result in our tcp buffer...
			# 
			global buffer$webcache_
			set data_ [set buffer$webcache_]
			set len [string length $data_]
			#puts "HIT $url $len"
			mc_trace_ul $url $len
			$self set_binary
			$self send_header $len $prefix
			$self send_data
			$self close

			delete $self
			return
		}

		#
		# otherwise, fetch the url from the origin server
		# XXX here's were we should do a distributed
		# SRM leader election to prevent synchronized
		# bombardment of the server
		#
		#$self launch $url
		$webcache_ request_doc $url $self
	}
}

ProxyConnection private progress args {
	puts -nonewline stderr .
	flush stderr
}

# Called from recv only when we aren't using the webcache (for timing
# without the cache).
ProxyConnection private direct_launch url  {
	$self instvar token_ url_
	set token_ [::http::geturl $url -command "$self direct_page_ready" ]
        set url_($token_) $url
}

# Called only from direct_page_ready
ProxyConnection private content_type token {
	upvar #0 $token state
	foreach {name value} $state(meta) {
                if {[regexp -nocase ^content-type$ $name]} {
			return [string trim $value]
                }
	}
	return ""
}

# Called only from direct_launch when we aren't using the webcache
# (for timing without the cache).
ProxyConnection private direct_page_ready token {
	set status [::http::status $token]
	if { $status != "ok" } {
		puts stderr "PAGE_READY ERROR $code"
	} else {
		set len [::http::size $token]
		set type [$self content_type $token]
	
		$self set_binary
		$self send_header $len $type

		$self instvar url_
		mc_trace_ul $url_($token) $len

		$self instvar data_
		set data_ [::http::data $token]
		$self send_data
	}
	
	$self close
	delete $self
}


ProxyConnection private send { s } {
    $self next $s
#    puts stderr "SEND: $s"
}


#
# Send a valid OK HTTP header to the attached client on the
# underlying TCP socket.
#
ProxyConnection private send_header { len content } {
#XXX each send does a flush
	$self send "HTTP/1.1 200 OK\n"
#	$self send "Date: [exec date -u]"
	$self send "Server: Mashcast Proxy\n"
	$self send "Connection: close\n"
# XXX need:
#Content-Type: text/html
#Last-Modified: Sun, 22 Dec 1996 05:46:43 GMT
#ETag: "1e125-364-32bccb43"

	$self send "Content-type: $content\n"
	$self send "Content-Length: $len\n"
	$self send "Accept-Ranges: bytes\n"
	$self send "\n"
}

ProxyConnection instproc upload {len type} {
#puts stderr "UPLOAD $len $type $self"
        #set len [::http::size $token]

	$self set_binary
	$self send_header $len $type
        #
	# XXX hack - otcl methods can't take binary strings, so stuff
	# the data in an instance variable and call send_data without args
        #
	$self instvar data_
	#set data_ [::http::data $token]
        #set data_ $data (doesn't work apparently)
        $self instvar webcache_
        global buffer$webcache_
	set data_ [set buffer$webcache_]
	$self send_data
	$self close
	#XXX
	delete $self
}

ProxyConnection instproc displayURL {doc} {
	exec  netscape -remote openFile($env(MC_CACHE_DIR)/$doc)
}









