import DaliSubprogram
import RealParameter

Class WhirlpoolScSubprogram -superclass DaliSubprogram

WhirlpoolScSubprogram instproc init {args} {
    eval $self next $args;
    
    # Set up inputs
    
    $self instvar input_id_list_;
    $self instvar input_info_;
    
    lappend input_id_list_ i1
    
    set input_info_(i1,spec) "";
    set input_info_(i1,trigger) 0;
    set input_info_(i1,buffertype) Uncompressed;
    set input_info_(i1,buffername) [new VidRep/Uncompressed];
    set input_info_(i1,decoder) "";
    
    # Set up outputs
    
    $self instvar output_id_list_;
    $self instvar output_info_;
    
    lappend output_id_list_ o1;
    
    set output_info_(o1,spec) "";
    set output_info_(o1,buffertype) Semicompressed;
    set output_info_(o1,buffername) [new VidRep/Semicompressed];
    set output_info_(o1,encoder) "";
    set output_info_(o1,format) SC;
    set output_info_(o1,vagent) "";
    set output_info_(o1,geometry) [list 0.0 0.0 1.0 1.0];
    set output_info_(o1,geometry_change) 0;

    # Set up parameters

    $self instvar parameter_id_list_;
    $self instvar parameter_info_;

    lappend parameter_id_list_ theta
    set pobj [new RealParameter];

    set parameter_info_(theta,oname) $pobj;
    $pobj from 0.0
    $pobj to 10.0
    $pobj set 0.0
    
}

WhirlpoolScSubprogram instproc set_output_geometry {id geometry} {
    $self instvar output_info_;

    set output_info_($id,geometry) $geometry;
    set output_info_(o1,geometry_change) 1;
}

WhirlpoolScSubprogram instproc trigger {} {
    $self instvar parameter_info_;
    $self instvar input_info_;
    $self instvar output_info_;
    $self instvar init_done_;
    $self instvar old_theta_;
    $self instvar la lb lc ld le lf ca cb cc cd ce cf
    $self instvar temp_out_;
    $self instvar tout_l_ tout_cr_ tout_cb_;

    set angle_obj $parameter_info_(theta,oname);
    set angle [$angle_obj get];
    
    set in_frame $input_info_(i1,buffername);
    set out_frame $output_info_(o1,buffername);
    
    if {![info exists init_done_]} {

	if {[$in_frame set w_] == 0} {
	    return;
	}

	set temp_out_ [new VidRep/Uncompressed];
	$temp_out_ copy_geometry $in_frame;
	$temp_out_ set h_subsample_ 2;
	$temp_out_ set v_subsample_ 1;

	puts "TempOut geometry: [$temp_out_ set w_] [$temp_out_ set h_] [$temp_out_ set true_w_] [$temp_out_ set true_h_] [$temp_out_ set h_subsample_] [$temp_out_ set v_subsample_]"
	$temp_out_ allocate;

	set x0 [lindex $output_info_(o1,geometry) 0];
	set y0 [lindex $output_info_(o1,geometry) 1];
	set x1 [lindex $output_info_(o1,geometry) 2];
	set y1 [lindex $output_info_(o1,geometry) 3];

	set xp0 [expr ((int([$in_frame set true_w_] * $x0) + 8) / 16) * 16];
	set xp1 [expr ((int([$in_frame set true_w_] * $x1) + 8) / 16) * 16];
	set yp0 [expr ((int([$in_frame set true_h_] * $y0) + 4) / 8) * 8];
	set yp1 [expr ((int([$in_frame set true_h_] * $y1) + 4) / 8) * 8];
	
	$out_frame set true_w_ [$in_frame set true_w_];
	$out_frame set true_h_ [$in_frame set true_h_];
	$out_frame set xoff_ $xp0;
	$out_frame set yoff_ $yp0;
	$out_frame set w_ [expr $xp1 - $xp0]
	$out_frame set h_ [expr $yp1 - $yp0]
	$out_frame set h_subsample_ 2;
	$out_frame set v_subsample_ 1;
	puts "OutFrame geometry: [$out_frame set w_] [$out_frame set h_] [$out_frame set true_w_] [$out_frame set true_h_] [$out_frame set h_subsample_] [$out_frame set v_subsample_]"
	$out_frame allocate;

	puts "TOut lum params: [$temp_out_ get_lum_name] $xp0 $yp0 [expr $xp1 - $xp0] [expr $yp1 - $yp0]"
	puts "TOut cr params: [$temp_out_ get_cr_name] [expr $xp0 / 2] $yp0 [expr ($xp1 - $xp0)/ 2] [expr $yp1 - $yp0]"
	puts "Tout cb params: [$temp_out_ get_cb_name] [expr $xp0 / 2] $yp0 [expr ($xp1 - $xp0)/ 2] [expr $yp1 - $yp0]"

	set tout_l_ [byte_clip [$temp_out_ get_lum_name] $xp0 $yp0 [expr $xp1 - $xp0] [expr $yp1 - $yp0]];
	set tout_cr_ [byte_clip [$temp_out_ get_cr_name] [expr $xp0 / 2] $yp0 [expr ($xp1 - $xp0)/ 2] [expr $yp1 - $yp0]];
	set tout_cb_ [byte_clip [$temp_out_ get_cb_name] [expr $xp0 / 2] $yp0 [expr ($xp1 - $xp0)/ 2] [expr $yp1 - $yp0]];

	set init_done_ 1;
	set old_theta_ "";
    }

    if {$output_info_(o1,geometry_change) == 1} {
	set x0 [lindex $output_info_(o1,geometry) 0];
	set y0 [lindex $output_info_(o1,geometry) 1];
	set x1 [lindex $output_info_(o1,geometry) 2];
	set y1 [lindex $output_info_(o1,geometry) 3];
	
	set xp0 [expr ((int([$in_frame set true_w_] * $x0) + 8) / 16) * 16];
	set xp1 [expr ((int([$in_frame set true_w_] * $x1) + 8) / 16) * 16];
	set yp0 [expr ((int([$in_frame set true_h_] * $y0) + 4) / 8) * 8];
	set yp1 [expr ((int([$in_frame set true_h_] * $y1) + 4) / 8) * 8];
	
	$out_frame set true_w_ [$in_frame set true_w_];
	$out_frame set true_h_ [$in_frame set true_h_];
	$out_frame set xoff_ $xp0;
	$out_frame set yoff_ $yp0;
	$out_frame set w_ [expr $xp1 - $xp0]
	$out_frame set h_ [expr $yp1 - $yp0]
	$out_frame set h_subsample_ 2;
	$out_frame set v_subsample_ 1;
	puts "OutFrame geometry change: [$out_frame set w_] [$out_frame set h_] [$out_frame set true_w_] [$out_frame set true_h_] [$out_frame set h_subsample_] [$out_frame set v_subsample_]"

	$out_frame allocate;

	byte_free $tout_l_;
	byte_free $tout_cr_;
	byte_free $tout_cb_;

	puts "TOut lum change: [$temp_out_ get_lum_name] $xp0 $yp0 [expr $xp1 - $xp0] [expr $yp1 - $yp0]"
	puts "TOut cr change: [$temp_out_ get_cr_name] [expr $xp0 / 2] $yp0 [expr ($xp1 - $xp0)/ 2] [expr $yp1 - $yp0]"
	puts "Tout cb change: [$temp_out_ get_cb_name] [expr $xp0 / 2] $yp0 [expr ($xp1 - $xp0)/ 2] [expr $yp1 - $yp0]"
	
	set tout_l_ [byte_clip [$temp_out_ get_lum_name] $xp0 $yp0 [expr $xp1 - $xp0] [expr $yp1 - $yp0]];
	set tout_cr_ [byte_clip [$temp_out_ get_cr_name] [expr $xp0 / 2] $yp0 [expr ($xp1 - $xp0)/ 2] [expr $yp1 - $yp0]];
	set tout_cb_ [byte_clip [$temp_out_ get_cb_name] [expr $xp0 / 2] $yp0 [expr ($xp1 - $xp0)/ 2] [expr $yp1 - $yp0]];
	
	set output_info_(o1,geometry_change) 0;
    }
    flush stdout
    
    if {$old_theta_ != $angle} {
	set w [$in_frame set w_];
	set h [$in_frame set h_];
    
	set cos_angle [expr cos($angle)];
	set sin_angle [expr sin($angle)];

	set in_h_sub [expr 1.0*[$in_frame set h_subsample_]]
	set in_v_sub [expr 1.0*[$in_frame set v_subsample_]]
	set out_h_sub [expr 1.0*[$out_frame set h_subsample_]]
	set out_v_sub [expr 1.0*[$out_frame set v_subsample_]]

	set sf [expr 1.0 - ($angle/10.0)];

	set la [expr $sf * $cos_angle];
	set lb [expr -($sf * $sin_angle)];
	set lc [expr ($w - ($sf * $w * $cos_angle) + ($sf * $h * $sin_angle)) / 2.0];
	set ld [expr $sf * $sin_angle];
	set le [expr $sf * $cos_angle];
	set lf [expr ($h - ($h * $sf * $cos_angle) - ($sf * $w * $sin_angle)) / 2.0];

	set ca [expr $la * $in_h_sub / $out_h_sub];
	set cb [expr $lb * $in_v_sub / $out_h_sub];
	set cc [expr $lc / $out_h_sub];
	set cd [expr $ld * $in_h_sub / $out_v_sub];
	set ce [expr $le * $in_v_sub / $out_v_sub];
	set cf [expr $lf / $out_v_sub];

	set old_theta_ $angle;
    }

    set in_l [$in_frame get_lum_name];
    set in_cr [$in_frame get_cr_name];
    set in_cb [$in_frame get_cb_name];

    byte_set $tout_l_ 0
    byte_set $tout_cr_ 128
    byte_set $tout_cb_ 128


    byte_affine $in_l $tout_l_ $la $lb $lc $ld $le $lf;
    byte_affine $in_cr $tout_cr_ $ca $cb $cc $cd $ce $cf;
    byte_affine $in_cb $tout_cb_ $ca $cb $cc $cd $ce $cf;

    set out_l [$out_frame get_lum_name];
    set out_cr [$out_frame get_cr_name];
    set out_cb [$out_frame get_cb_name];

    
    $out_frame set ts_ [$in_frame set ts_];
    byte_to_sc $tout_l_ $out_l 0;
    byte_to_sc $tout_cr_ $out_cr 1;
    byte_to_sc $tout_cb_ $out_cb 1;

    set encoder $output_info_(o1,encoder);

    if {$encoder != ""} {
	$encoder recv $out_frame;
    }

    $self send_completion_token
}



