
import GraphComm

Class ScSpatialCombiner

ScSpatialCombiner instproc init {id cntrl_spec} {
    $self next;

    $self instvar output_id_list_;
    $self instvar output_info_;

    $self instvar subgraph_list_;
    $self instvar subgraph_info_;

    $self instvar sid_;
    
    $self instvar id_;
    $self instvar comm_obj_;

    $self instvar callback_array_;

    set output_id_list_ "";

    set id_ $id;

    set cntrl_spec [split $cntrl_spec "/"];
    set addr [lindex $cntrl_spec 0];
    set port [lindex $cntrl_spec 1];
    set ttl [lindex $cntrl_spec 2];

    set comm_obj_ [new GraphComm/ScSpatialCombiner $self $id_ $addr $port $ttl];

    set sid_ -1;
    set subgraph_list_ "";
}

ScSpatialCombiner instproc add_subgraph {new_subgraph} {
    $self instvar subgraph_info_ subgraph_list_ sid_;
    $self instvar comm_obj_ id_

    foreach sid $subgraph_list_ {
	if {$subgraph_info_($sid,spec) == $new_subgraph} {
	    return;
	}
    }

    incr sid_;

    lappend subgraph_list_ $sid_;

    set subgraph_info_($sid_,spec) $new_subgraph;
    set subgraph_info_($sid_,comm_obj) [new GraphComm/ScCombToSubgraph $self $id_ $new_subgraph];

    
    # Map trigger commands from above.

    $subgraph_info_($sid_,comm_obj) send_map_command [list map_trigger_cmds [$comm_obj_ primary_addr] [$comm_obj_ primary_port] [$comm_obj_ primary_ttl]];
    
    # Map inputs both up and down.

    $subgraph_info_($sid_,comm_obj) send_map_command [list map_inputs [$comm_obj_ primary_addr] [$comm_obj_ primary_port] [$comm_obj_ primary_ttl]];
    
    $comm_obj_ send_map_command [list map_inputs [$subgraph_info_($sid_,comm_obj) primary_addr] [$subgraph_info_($sid_,comm_obj) primary_port] [$subgraph_info_($sid_,comm_obj) primary_ttl]];

    # Map parameters up and down

    $subgraph_info_($sid_,comm_obj) send_map_command [list map_parameters [$comm_obj_ primary_addr] [$comm_obj_ primary_port] [$comm_obj_ primary_ttl]];
    
    $comm_obj_ send_map_command [list map_parameters [$subgraph_info_($sid_,comm_obj) primary_addr] [$subgraph_info_($sid_,comm_obj) primary_port] [$subgraph_info_($sid_,comm_obj) primary_ttl]];

    # Recalculate any existing divisions
    $self recalculate_division

}

ScSpatialCombiner instproc del_subgraph {subgraph_to_del} {
    $self instvar subgraph_list_ subgraph_info_;

    set sid_to_del "";
    foreach s $subgraph_list_ {
	if {$subgraph_info_($s,spec) == $subgraph_to_del} {
	    set sid_to_del $s;
	    break;
	}
    }
    if {$sid_to_del == ""} return;
    
    delete $subgraph_info_($sid_to_del,comm_obj);
    unset subgraph_info_($sid_to_del,comm_obj);
    unset subgraph_info_($sid_to_del,spec);

    set idx [lsearch $subgraph_list_ $sid_to_del];
    set subgraph_list_ [lreplace $subgraph_list_ $idx $idx];

    $self recalculate_division;
}

ScSpatialCombiner instproc recalculate_division {} {
    $self instvar subgraph_list_;
    $self instvar subgraph_info_;
    $self instvar output_id_list_;

    set num_subgraph [llength $subgraph_list_];

    set x0 0.0
    set x1 1.0

    foreach s $subgraph_list_ {
	set idx [lsearch $subgraph_list_ $s];
	set comm_obj $subgraph_info_($s,comm_obj);

	set y0 [expr ($idx * 1.0) / ($num_subgraph * 1.0)];
	set y1 [expr ($idx * 1.0 + 1.0) / ($num_subgraph * 1.0)];

	foreach o $output_id_list_ {
	    $comm_obj set_output_attr $o geometry [list $x0 $y0 $x1 $y1];
	}
    }
}

ScSpatialCombiner instproc new_output {subgraph_comm o} {

    $self instvar output_id_list_ subgraph_list_
    $self instvar subgraph_info_ output_info_

    if {[lsearch $output_id_list_ $o] == -1} {
	lappend output_id_list_ $o;

	# Brand new output discovered.

	set spec [$self GenerateSpec];
	set spec_split [split $spec "/"];
	set addr [lindex $spec_split 0];
	set port [lindex $spec_split 1];
	
	set output_info_($o,combiner_obj) [new Module/Combine/Spatial/SC];
	set vagent_array_($addr,$port) [new ScSpatialCombinerVideoAgent $addr/$port $output_info_($o,combiner_obj)];
	set output_info_($o,vagent) $vagent_array_($addr,$port);
	set output_info_($o,in_spec) $addr/$port;
	set output_info_($o,out_vagent) "";
	
	set output_info_($o,sc_decoder) [new Module/VideoDecoder/SCToSemicompressed];
	set output_info_($o,sc_buffer) [new VidRep/Semicompressed];
	$output_info_($o,sc_decoder) set_frame_buffer $output_info_($o,sc_buffer);

	set output_info_($o,encoder) [new Module/VideoEncoder/SemicompressedToJPEG];
	set output_info_($o,bufferpool) [new BufferPool/RTP];
	$output_info_($o,encoder) buffer-pool $output_info_($o,bufferpool);
	$output_info_($o,encoder) mtu [$self get_option mtu];
	$output_info_($o,sc_decoder) set_callback "$output_info_($o,encoder) recv $output_info_($o,sc_buffer)"
	
	$output_info_($o,combiner_obj) target $output_info_($o,sc_decoder);

	# Reflect new output in our control session

	$self instvar comm_obj_;

	$comm_obj_ create_output $o;
    }
	
    $subgraph_comm create_output $o;
    $subgraph_comm create_output_attr $o spec;
    $subgraph_comm create_output_attr $o geometry;
    $subgraph_comm set_output_attr $o spec $output_info_($o,in_spec);

    foreach s $subgraph_list_ {
	if {$subgraph_info_($s,comm_obj) == $subgraph_comm} {
	    break;
	}
    }
    set idx [lindex $subgraph_list_ $s];
    set num_subgraph [llength $subgraph_list_];
    
    set x0 0.0;
    set x1 1.0;
    set y0 [expr ($idx * 1.0) / ($num_subgraph * 1.0)];
    set y1 [expr ($idx * 1.0 + 1.0) / ($num_subgraph * 1.0)];
    
    $subgraph_comm set_output_attr $o geometry [list $x0 $y0 $x1 $y1];
}

ScSpatialCombiner instproc set_output_spec {out_id spec} {
    $self instvar output_id_list_ output_info_

    if {[lsearch $output_id_list_ $out_id] == -1} {
	return;
    }

    set combiner_obj $output_info_($out_id,combiner_obj);

    set spec_split [split $spec "/"];
    
    set addr [lindex $spec_split 0];
    set port [lindex $spec_split 1];
    
    set output_info_($out_id,out_vagent) [new VideoAgent $self $addr/$port];
    $output_info_($out_id,out_vagent) local_bandwidth 30000000;

    $combiner_obj ssrc [$output_info_($out_id,out_vagent) get_local_srcid];
    $output_info_($out_id,bufferpool) srcid [$output_info_($out_id,out_vagent) get_local_srcid];
    $output_info_($out_id,encoder) target [$output_info_($out_id,out_vagent) get_transmitter];
}

ScSpatialCombiner instproc GenerateSpec {} {
    # OK, this is a REAL hack

    global spec_generator;

    if {![info exists spec_generator(init)]} {
	set spec_generator(init) 1;
	set spec_generator(b1) 224;
	set spec_generator(b2) 5;
	set spec_generator(b3) 9;
	set spec_generator(b4) 1;
	set spec_generator(port) 22334;
    }

    incr spec_generator(b4);

    return "$spec_generator(b1).$spec_generator(b2).$spec_generator(b3).$spec_generator(b4)/$spec_generator(port)/16";
}
    

Class GraphComm/ScSpatialCombiner -superclass GraphComm

GraphComm/ScSpatialCombiner instproc init {sc_combiner id addr port ttl} {
    $self next $id $addr $port $ttl;

    $self instvar sc_combiner_;

    set sc_combiner_ $sc_combiner;
}


GraphComm/ScSpatialCombiner instproc recv_misc {data} {
    $self next $data;

    $self instvar sc_combiner_;

    set cmd [lindex $data 0];

    switch -exact -- $cmd {
	add_subgraph {
	    $sc_combiner_ add_subgraph [lindex $data 1];
	}
	del_subgraph {
	    $sc_combiner_ del_subgraph [lindex $data 1];
	}
    }
}

GraphComm/ScSpatialCombiner instproc update_output_attr_value {output_name attr_name value} {
    $self next $output_name $attr_name $value;

    if {$attr_name == "spec"} {
	$self instvar sc_combiner_;

	$sc_combiner_ set_output_spec $output_name $value;
    }
}

Class GraphComm/ScCombToSubgraph -superclass GraphComm

GraphComm/ScCombToSubgraph instproc init {sc_combiner id spec} {
    set spec [split $spec "/"];
    set addr [lindex $spec 0];
    set port [lindex $spec 1];
    set ttl [lindex $spec 2];

    $self next $id $addr $port $ttl;

    $self instvar sc_combiner_;

    set sc_combiner_ $sc_combiner;
}

GraphComm/ScCombToSubgraph instproc new_output {new_name} {
    $self next $new_name;

    $self instvar sc_combiner_;

    $sc_combiner_ new_output $self $new_name;
}

import VideoAgent

Class ScSpatialCombinerVideoAgent -superclass VideoAgent

ScSpatialCombinerVideoAgent instproc init {spec combiner} {
    eval $self next $self $spec;

    $self instvar combiner_;

    set combiner_ $combiner;
}

ScSpatialCombinerVideoAgent instproc create_decoder src {
    $self instvar combiner_;

    set d [$src handler];
    if {$d != ""} {
	delete $d;
    }

    return $combiner_;
}

ScSpatialCombinerVideoAgent instproc deactivate src {
    puts "$src has been deactivated!!!!";
}
