/*********************************************************************
 * This Software is copyright INRIA. 1997.
 * 
 * INRIA holds all the ownership rights on the Software. The scientific
 * community is asked to use the SOFTWARE in order to test and evaluate
 * it.
 * 
 * INRIA freely grants the right to use the Software. Any use or
 * reproduction of this Software to obtain profit or for commercial ends
 * being subject to obtaining the prior express authorization of INRIA.
 * 
 * INRIA authorizes any reproduction of this Software
 * 
 *  - in limits defined in clauses 9 and 10 of the Berne agreement for
 * the protection of literary and artistic works respectively specify in
 * their paragraphs 2 and 3 authorizing only the reproduction and quoting
 * of works on the condition that :
 * 
 *  - "this reproduction does not adversely affect the normal
 * exploitation of the work or cause any unjustified prejudice to the
 * legitimate interests of the author".
 * 
 *  - that the quotations given by way of illustration and/or tuition
 * conform to the proper uses and that it mentions the source and name of
 * the author if this name features in the source",
 * 
 *  - under the condition that this file is included with any reproduction.
 * 
 * Any commercial use made without obtaining the prior express agreement
 * of INRIA would therefore constitute a fraudulent imitation.
 * 
 * The Software being currently developed, INRIA is assuming no
 * liability, and should not be responsible, in any manner or any case,
 * for any direct or indirect damages sustained by the user.
 * 
 * Furthermore, INRIA grants for free the right to adapt, modify or integrate
 * the SOFTWARE in another program, provided that any derived work 
 * should be distributed in the same conditions as the SOFTWARE. 
 *********************************************************************
 */

/*
 * Filename: RemoteController.java
 *   -- Created on Thu Aug 13 1998
 *   -- Author: Yatin Chawathe <yatin@cs.berkeley.edu>
 */


/*
 * Client program running in a Java-enabled browser. It just
 * receives data from multicast deamon and displays in browser's
 * right frame.
 */

import java.awt.*;
import java.applet.*;
import java.net.*;
import java.util.*;
import java.io.*;


public class RemoteController extends Applet implements Runnable {
	private String host_;
	private int port_;
	private Thread thread_;
	private Socket sock_;
	private BufferedReader input_;
	private String status_;
	private Exception exception_;
	
public void init() {
	URL url;
	url = getDocumentBase();
	host_ = url.getHost();

	/* get the port number to connect to either from the param list,
	 * or from the URL itself
	 */
	String port_str=null, file = url.getFile();
	int idx = file.indexOf('?');
	if (idx >= 0) {
		port_str = file.substring(idx+1);
	}
	System.out.println("FFFFFFFFFFFFFFF: |" + port_str + "|");
	if (port_str==null) port_str = getParameter("applet_port");
	if (port_str!=null) {
		try { port_ = Integer.parseInt(port_str, 10); }
		catch (Exception e) { port_ = url.getPort(); }
	} else port_ = url.getPort();

	System.out.println("URL: " + url);
	System.out.println("Protocol: " + url.getProtocol());
	System.out.println("Host: " + url.getHost());
	System.out.println("Port: " + port_);
	System.out.println("File: " + url.getFile());
	System.out.println("Ref: " + url.getRef());
}

	
public void start() {
	System.out.println("starting...");

	try {
	    sock_ = new Socket(host_, port_);
	    input_ = new BufferedReader(new
					InputStreamReader(sock_.
							  getInputStream()));
	} catch (IOException e) {
	    status("Could not connect to " + host_ + ":" + port_, e);
	    System.out.println("Could not connect to " + host_ + ":" + port_
			       + "\n" + e);
	    return;
	}

	System.out.println("Successfully connected to " + host_ + ":" + port_
			   + "\n");
	thread_ = new Thread(this);
	thread_.start();
}


public void stop() {
	System.out.println("stopped.");

	if (sock_ != null) {
		try {
			sock_.close();
		} catch (Exception e) {
			System.out.println("RemoteController: close() error: "
					   + e);
		}
	}
	sock_ = null;
	thread_.destroy();
	thread_ = null;
}


public void run() {
	String data;

	while (thread_ != null) {
		try {
			data = input_.readLine();
		} catch (Exception e) {
			lost();
			return;
		}
		
		System.out.println("req " + data);
		if (data==null) {
			lost();
			return;
		}

		if (data.startsWith("show:")) {
			show_document(data.substring(5, data.length()));
		} else {
			System.out.println("Unknown req: " + data);
		}
	}
}


private void lost() {
	stop();
	status("ERROR: applet connection lost", null);
}

	
private void status(String s, Exception e) {
	getAppletContext().showStatus(s);
	status_ = s;
	exception_ = e;
	repaint();
}


public void paint(Graphics g) {
	if (status_!=null) {
		int xoff = 2;
		int yoff = g.getFontMetrics().getHeight()/2 + 2;
		g.drawString(status_, xoff, yoff);
	}

	if (exception_!=null) {
		int xoff = 2;
		int yoff = (g.getFontMetrics().getHeight()*3)/2 + 2;
		g.drawString(exception_.toString(), xoff, yoff);
	}
}


private void show_document(String url) {
	try {
		status("showing document " + url, null);
		getAppletContext().showDocument(new URL(url),
						getParameter("target"));
	} catch (MalformedURLException e) {
		System.out.println("Invalid URL" + e);
	}
}

	
}
